/* ccgo: igs/control.cc
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "control.hh"
#include "parser/init.hh"
#include <sstream>
#include <iostream>

using namespace igs;

void Control::process_line()
{
	if (get_line() == "") return;
	num_cmd = -1;
	if (parser) {
		if (parser->process()) return;
		delete parser;
		parser = 0;
	}
	parser = Parser::search_id(this);
}

void Control::queue_cmd(const std::string & c)
{
	cmd_queue.push(c);
}

void Control::queue_process()
{
	if (! cmd_queue.empty()) {
		send_line(cmd_queue.front());
		cmd_queue.pop();
	}
}

Control::Control() :
	parser(0)
{
}

Control::~Control()
{
	if (parser) delete parser;
}

const std::string & Control::get_line()
{
	return line_buffer;
}

bool Control::open(const std::string & n)
{
// 	std::cerr << "Control::open" << std::endl;
	if (! open_conn(n)) return false;
	set_no_wait(true);
	parser::Init * i = new parser::Init(this);
	parser = i;
	// i->set_login("john", "");
	return true;
}

void Control::close()
{
	close_conn();
}

int Control::get_num_cmd()
{
	if (num_cmd < 0) {
		std::istringstream s(get_line());
		s >> num_cmd;
		if (num_cmd < 0) num_cmd = 0;
		else {
			s.get();
			int pos = get_line().length() - s.rdbuf()->in_avail();
			cmd_dat = get_line().substr(pos);
			// std::cerr << "spliting line: (" << num_cmd << ")" << cmd_dat << std::endl;
		}
	}
	return num_cmd;
}

const std::string & Control::get_cmd_dat()
{
	return cmd_dat;
}

void Control::client_ready()
{
}

void Control::set_prompt(PromptType p)
{
}

void Control::set_username(const std::string &)
{
}

void Control::do_beep()
{
}

void Control::add_message(const std::string &)
{
}

void Control::add_tagged_message(const std::string & tag, const std::string & text)
{
	// std::cerr << "add tagged" << std::endl;
}

void Control::player_update(const Player &)
{
}

void Control::player_stats(const Player &)
{
}

void Control::player_odds(const Odds &)
{
}

void Control::player_end_update()
{
}

void Control::game_result(const std::vector<Result> &)
{
}

void Control::stored_game(const std::vector<Storage> &)
{
}

void Control::game_update(const Game &)
{
}

void Control::game_end_update()
{
}

void Control::game_move(go::Move * mv, int g)
{
	std::cerr << "game " << g << ": " << mv->as_string() << std::endl;
}

void Control::game_undo(int)
{
}

void Control::game_time(int g, int wt, int wm, int bt, int bm)
{
}

void Control::ob_remove(int)
{
}

void Control::game_finish(int, const std::string &)
{
}

void Control::match_score(int)
{
}

void Control::match_restore(int)
{
}

void Control::match_remove(const go::Loc &, int)
{
}

void Control::match_adjourn(int)
{
}

void Control::match_finish(const std::string &, int)
{
}

void Control::match_resign(const std::string &)
{
}

void Control::match_timeup(const std::string &)
{
}

void Control::remove_match(const std::string &, const std::string &)
{
}

void Control::player_update(const std::string &, const std::string &)
{
}

void Control::player_remove(const std::string &)
{
}

void Control::message_sys(const std::string &)
{
}

void Control::message_tell(const std::string &)
{
}

void Control::message_tell(const std::string &, const std::string &)
{
}

void Control::message_yell(const std::string &)
{
}

void Control::message_kibitz(int, const std::string &)
{
}

void Control::message_misc(const std::string &)
{
}

void Control::match_request(const Match &)
{
}

void Control::match_create(int, const std::string &)
{
}

void Control::match_decline(const std::string &)
{
}

void Control::game_board(const Board &)
{
}

void Control::toggle_state(ToggleType, bool)
{
}
